# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from .. import utils
from ..data_types.array import Array
from ..data_types.enums import Enums
from ..data_types.integer import Int
from ..data_types.decimal import Decimal
from .command_packet import CommandPacket
from ..enums.bia_enums import BIASweepFrequency, BIAAppInfo


class ADPDDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.ADPD6: ['0xC2', '0x16']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x2C',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'data_type': 4,
                'channel_num': 2,
                'timestamp': 1751731879,
                'sample_num': 2,
                'signal_data': [ 351762, 351755 ],
                'dark_data': [ 0, 0 ],
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["data_type"] = Int(2)
        self._config["payload"]["channel_num"] = Int(1)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["sample_num"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=1, data_types=[Int(1)])

    def get_dict(self, last_timestamp=None):
        """
        Reorganising of stream data in stream_data key.
        """
        packet = super().get_dict()
        packet["payload"]["dark_data"] = []
        packet["payload"]["signal_data"] = []
        if packet["payload"]["data_type"] & 0x100 == 0:
            signal_size = (packet["payload"]["data_type"] & 0xF)
            dark_size = (packet["payload"]["data_type"] & 0xF0) >> 4
            data_size = signal_size + dark_size
            sample_num = packet["payload"]["sample_num"]
            for i in range(0, sample_num * data_size, data_size):
                data = packet["payload"]["data"][i: i + dark_size]
                dark_data = utils.join_multi_length_packets(data)
                data = packet["payload"]["data"][i + dark_size: i + dark_size + signal_size]
                signal_data = utils.join_multi_length_packets(data)
                packet["payload"]["dark_data"].append(dark_data)
                packet["payload"]["signal_data"].append(signal_data)
        else:
            signal_size = (packet["payload"]["data_type"] & 0xFF)
            for i in range(0, signal_size, 2):
                packet["payload"]["dark_data"].append(0)
                data = packet["payload"]["data"][i: i + 1]
                signal_data = utils.join_multi_length_packets(data)
                packet["payload"]["signal_data"].append(signal_data)
        del packet["payload"]["data"]
        utils.update_timestamp(packet, last_timestamp)
        return packet


class ADXLDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.ADXL: ['0xC2', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x37',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 395,
                'data_type': 0,
                'stream_data': [
                    {
                        'timestamp': 1767361372,
                        'x': -85,
                        'y': 90,
                        'z': 55
                    },
                    {
                        'timestamp': 1767362027,
                        'x': -80,
                        'y': 88,
                        'z': 55
                    },
                    {
                        'timestamp': 1767362682,
                        'x': 40,
                        'y': 274,
                        'z': 79
                    },
                    {
                        'timestamp': 1767363337,
                        'x': 70,
                        'y': 273,
                        'z': 54
                    },
                    {
                        'timestamp': 1767363931,
                        'x': 57,
                        'y': 257,
                        'z': 48
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["data_type"] = Int(1)
        self._config["payload"]["timestamp1"] = Int(4)
        self._config["payload"]["x1"] = Int(2, sign=True)
        self._config["payload"]["y1"] = Int(2, sign=True)
        self._config["payload"]["z1"] = Int(2, sign=True)
        self._config["payload"]["timestamp2"] = Int(2)
        self._config["payload"]["x2"] = Int(2, sign=True)
        self._config["payload"]["y2"] = Int(2, sign=True)
        self._config["payload"]["z2"] = Int(2, sign=True)
        self._config["payload"]["timestamp3"] = Int(2)
        self._config["payload"]["x3"] = Int(2, sign=True)
        self._config["payload"]["y3"] = Int(2, sign=True)
        self._config["payload"]["z3"] = Int(2, sign=True)
        self._config["payload"]["timestamp4"] = Int(2)
        self._config["payload"]["x4"] = Int(2, sign=True)
        self._config["payload"]["y4"] = Int(2, sign=True)
        self._config["payload"]["z4"] = Int(2, sign=True)
        self._config["payload"]["timestamp5"] = Int(2)
        self._config["payload"]["x5"] = Int(2, sign=True)
        self._config["payload"]["y5"] = Int(2, sign=True)
        self._config["payload"]["z5"] = Int(2, sign=True)

    def get_dict(self, last_timestamp=None):
        """
        Reorganising of stream data in stream_data key.
        """
        packet = super().get_dict()
        packet["payload"]["stream_data"] = []
        timestamp = 0
        for i in range(1, 6):
            timestamp += packet["payload"][f"timestamp{i}"]
            data = {
                "timestamp": timestamp,
                "x": packet["payload"][f"x{i}"],
                "y": packet["payload"][f"y{i}"],
                "z": packet["payload"][f"z{i}"]
            }
            [packet["payload"].pop(key) for key in [f"timestamp{i}", f"x{i}", f"y{i}", f"z{i}"]]
            packet["payload"]["stream_data"].append(data)
        utils.update_timestamp(packet, last_timestamp)
        return packet


class ECGDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.ECG: ['0xC4', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x3D',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'data_type': 1,
                'ecg_info': 0,
                'hr': 0,
                'stream_data': [
                    {
                        'timestamp': 1771095376,
                        'ecg_data': 12351
                    },
                    {
                        'timestamp': 1771095703,
                        'ecg_data': 12353
                    },
                    {
                        'timestamp': 1771096030,
                        'ecg_data': 52470
                    },
                    {
                        'timestamp': 1771096357,
                        'ecg_data': 41129
                    },
                    {
                        'timestamp': 1771096676,
                        'ecg_data': 63838
                    },
                    {
                        'timestamp': 1771096995,
                        'ecg_data': 63848
                    },
                    {
                        'timestamp': 1771097314,
                        'ecg_data': 63848
                    },
                    {
                        'timestamp': 1771097633,
                        'ecg_data': 63848
                    },
                    {
                        'timestamp': 1771097954,
                        'ecg_data': 63833
                    },
                    {
                        'timestamp': 1771098273,
                        'ecg_data': 63846
                    },
                    {
                        'timestamp': 1771098592,
                        'ecg_data': 63846
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["data_type"] = Int(1)
        self._config["payload"]["timestamp1"] = Int(4)
        self._config["payload"]["ecg_info"] = Int(1)
        self._config["payload"]["hr"] = Int(1)
        self._config["payload"]["ecg_data1"] = Int(2)
        self._config["payload"]["timestamp2"] = Int(2)
        self._config["payload"]["ecg_data2"] = Int(2)
        self._config["payload"]["timestamp3"] = Int(2)
        self._config["payload"]["ecg_data3"] = Int(2)
        self._config["payload"]["timestamp4"] = Int(2)
        self._config["payload"]["ecg_data4"] = Int(2)
        self._config["payload"]["timestamp5"] = Int(2)
        self._config["payload"]["ecg_data5"] = Int(2)
        self._config["payload"]["timestamp6"] = Int(2)
        self._config["payload"]["ecg_data6"] = Int(2)
        self._config["payload"]["timestamp7"] = Int(2)
        self._config["payload"]["ecg_data7"] = Int(2)
        self._config["payload"]["timestamp8"] = Int(2)
        self._config["payload"]["ecg_data8"] = Int(2)
        self._config["payload"]["timestamp9"] = Int(2)
        self._config["payload"]["ecg_data9"] = Int(2)
        self._config["payload"]["timestamp10"] = Int(2)
        self._config["payload"]["ecg_data10"] = Int(2)
        self._config["payload"]["timestamp11"] = Int(2)
        self._config["payload"]["ecg_data11"] = Int(2)

    def get_dict(self, last_timestamp=None):
        """
        Reorganising of stream data in stream_data key.
        """
        packet = super().get_dict()
        packet["payload"]["stream_data"] = []
        timestamp = 0
        for i in range(1, 12):
            timestamp += packet["payload"][f"timestamp{i}"]
            data = {"timestamp": timestamp,
                    "ecg_data": packet["payload"][f"ecg_data{i}"]}
            [packet["payload"].pop(key) for key in [f"timestamp{i}", f"ecg_data{i}"]]
            packet["payload"]["stream_data"].append(data)
        utils.update_timestamp(packet, last_timestamp)
        return packet


class EDADataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.EDA: ['0xC4', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x3D',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'data_type': 0,
                'stream_data': [
                    {
                        'timestamp': 1774366622,
                        'real': 0,
                        'imaginary': 0
                    },
                    {
                        'timestamp': 1774374407,
                        'real': 0,
                        'imaginary': 16708
                    },
                    {
                        'timestamp': 1774382157,
                        'real': 16728,
                        'imaginary': 3257
                    },
                    {
                        'timestamp': 1774389924,
                        'real': 3277,
                        'imaginary': -20751
                    },
                    {
                        'timestamp': 1774397691,
                        'real': -20731,
                        'imaginary': -15161
                    },
                    {
                        'timestamp': 1774405458,
                        'real': -15141,
                        'imaginary': -30319
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["data_type"] = Int(1)
        self._config["payload"]["timestamp1"] = Int(4)
        self._config["payload"]["real1"] = Int(2, sign=True)
        self._config["payload"]["imaginary1"] = Int(2, sign=True)
        self._config["payload"]["timestamp2"] = Int(4)
        self._config["payload"]["real2"] = Int(2, sign=True)
        self._config["payload"]["imaginary2"] = Int(2, sign=True)
        self._config["payload"]["timestamp3"] = Int(4)
        self._config["payload"]["real3"] = Int(2, sign=True)
        self._config["payload"]["imaginary3"] = Int(2, sign=True)
        self._config["payload"]["timestamp4"] = Int(4)
        self._config["payload"]["real4"] = Int(2, sign=True)
        self._config["payload"]["imaginary4"] = Int(2, sign=True)
        self._config["payload"]["timestamp5"] = Int(4)
        self._config["payload"]["real5"] = Int(2, sign=True)
        self._config["payload"]["imaginary5"] = Int(2, sign=True)
        self._config["payload"]["timestamp6"] = Int(4)
        self._config["payload"]["real6"] = Int(2, sign=True)
        self._config["payload"]["imaginary6"] = Int(2, sign=True)

    def get_dict(self, last_timestamp=None):
        """
        Reorganising of stream data in stream_data key.
        """
        packet = super().get_dict()
        packet["payload"]["stream_data"] = []
        for i in range(1, 7):
            timestamp = packet["payload"][f"timestamp{i}"]
            data = {"timestamp": timestamp,
                    "real": packet["payload"][f"real{i}"],
                    "imaginary": packet["payload"][f"imaginary{i}"]}
            [packet["payload"].pop(key) for key in [f"timestamp{i}", f"real{i}", f"imaginary{i}"]]
            packet["payload"]["stream_data"].append(data)
        utils.update_timestamp(packet, last_timestamp)
        return packet


class PedometerDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.PEDOMETER: ['0xC4', '0x04']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x17',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 1,
                'steps': 0,
                'algo_status': 8193,
                'timestamp': 1776836792,
                'reserved': 0
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["steps"] = Int(4, sign=True)
        self._config["payload"]["algo_status"] = Int(2)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["reserved"] = Int(1)

    def get_dict(self, last_timestamp=None):
        """
        Postprocessing.
        """
        packet = super().get_dict()
        utils.update_timestamp(packet, last_timestamp)
        return packet


class TemperatureDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.TEMPERATURE: ['0xC4', '0x06']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x14',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 2,
                'timestamp': 1779692742,
                'skin_temperature': 30.1, # celsius
                'impedance': 79000 # ohm
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["skin_temperature"] = Int(2)
        self._config["payload"]["impedance"] = Int(2)

    def get_dict(self, last_timestamp=None):
        """
        Postprocessing.
        """
        packet = super().get_dict()
        packet["payload"]["skin_temperature"] = packet["payload"]["skin_temperature"] / 1000.0
        packet["payload"]["impedance"] = packet["payload"]["impedance"] * 100
        utils.update_timestamp(packet, last_timestamp)
        return packet


class PPGDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.PPG: ['0xC4', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x2E',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.NEW_STREAM_STATUS: ['0x43']>,
                'sequence_number': 9,
                'timestamp': 1782357951,
                'adpd_lib_state': 7,
                'hr': 0,
                'confidence': 0,
                'hr_type': 0,
                'rr_interval': 0,
                'debug_info': [ 0, 320, 50, 0, 5, 0, 320, 50, 978 ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["adpd_lib_state"] = Int(2)
        self._config["payload"]["hr"] = Int(2)
        self._config["payload"]["confidence"] = Int(2)
        self._config["payload"]["hr_type"] = Int(2)
        self._config["payload"]["rr_interval"] = Int(2)
        self._config["payload"]["debug_info"] = Array(-1, dimension=1, data_types=[Int(2)])

    def get_dict(self, last_timestamp=None):
        """
        Postprocessing.
        """
        packet = super().get_dict()
        utils.update_timestamp(packet, last_timestamp)
        return packet


class SYNCPPGDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.SYNC_PPG: ['0xC4', '0x05']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x48',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 119,
                'stream_data': [
                    {
                        'ppg_timestamp': 1782379042,
                        'ppg_data': 186278,
                        'adxl_timestamp': 1782378429,
                        'adxl_x': 48,
                        'adxl_y': 268,
                        'adxl_z': 62
                    },
                    {
                        'ppg_timestamp': 1782379682,
                        'ppg_data': 186278,
                        'adxl_timestamp': 1782379070,
                        'adxl_x': 55,
                        'adxl_y': 264,
                        'adxl_z': 60
                    },
                    {
                        'ppg_timestamp': 1782380323,
                        'ppg_data': 186289,
                        'adxl_timestamp': 1782379709,
                        'adxl_x': 68,
                        'adxl_y': 268,
                        'adxl_z': 57
                    },
                    {
                        'ppg_timestamp': 1782380962,
                        'ppg_data': 186305,
                        'adxl_timestamp': 1782380349,
                        'adxl_x': 60,
                        'adxl_y': 263,
                        'adxl_z': 62
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["ppg_timestamp1"] = Int(4)
        self._config["payload"]["adxl_timestamp1"] = Int(4)
        self._config["payload"]["ppg_timestamp2"] = Int(2)
        self._config["payload"]["ppg_timestamp3"] = Int(2)
        self._config["payload"]["ppg_timestamp4"] = Int(2)
        self._config["payload"]["adxl_timestamp2"] = Int(2)
        self._config["payload"]["adxl_timestamp3"] = Int(2)
        self._config["payload"]["adxl_timestamp4"] = Int(2)
        self._config["payload"]["ppg_data1"] = Int(4)
        self._config["payload"]["ppg_data2"] = Int(4)
        self._config["payload"]["ppg_data3"] = Int(4)
        self._config["payload"]["ppg_data4"] = Int(4)
        self._config["payload"]["adxl_x1"] = Int(2, sign=True)
        self._config["payload"]["adxl_x2"] = Int(2, sign=True)
        self._config["payload"]["adxl_x3"] = Int(2, sign=True)
        self._config["payload"]["adxl_x4"] = Int(2, sign=True)
        self._config["payload"]["adxl_y1"] = Int(2, sign=True)
        self._config["payload"]["adxl_y2"] = Int(2, sign=True)
        self._config["payload"]["adxl_y3"] = Int(2, sign=True)
        self._config["payload"]["adxl_y4"] = Int(2, sign=True)
        self._config["payload"]["adxl_z1"] = Int(2, sign=True)
        self._config["payload"]["adxl_z2"] = Int(2, sign=True)
        self._config["payload"]["adxl_z3"] = Int(2, sign=True)
        self._config["payload"]["adxl_z4"] = Int(2, sign=True)

    def get_dict(self, last_timestamp=None):
        """
        Postprocessing.
        """
        packet = super().get_dict()
        packet["payload"]["stream_data"] = []
        ppg_timestamp = 0
        adxl_timestamp = 0
        for i in range(1, 5):
            ppg_timestamp += packet["payload"][f"ppg_timestamp{i}"]
            adxl_timestamp += packet["payload"][f"adxl_timestamp{i}"]
            data = {"ppg_timestamp": ppg_timestamp,
                    "ppg_data": packet["payload"][f"ppg_data{i}"],
                    "adxl_timestamp": adxl_timestamp,
                    "adxl_x": packet["payload"][f"adxl_x{i}"],
                    "adxl_y": packet["payload"][f"adxl_y{i}"],
                    "adxl_z": packet["payload"][f"adxl_z{i}"]}
            [packet["payload"].pop(key) for key in
             [f"ppg_timestamp{i}", f"ppg_data{i}", f"adxl_timestamp{i}", f"adxl_x{i}", f"adxl_y{i}", f"adxl_z{i}"]]
            packet["payload"]["stream_data"].append(data)
        utils.update_timestamp(packet, last_timestamp, is_syncppg=True)
        return packet


class SQIDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.SQI: ['0xC8', '0x0D']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x19',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 0,
                'sqi': 2.537532566293521e-07,
                'sqi_slot': 42405,
                'algo_status': 0,
                'timestamp': 1786433199,
                'reserved': 0
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["sqi"] = Decimal(4)
        self._config["payload"]["sqi_slot"] = Int(2)
        self._config["payload"]["algo_status"] = Int(2)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["reserved"] = Int(1)

    def get_dict(self, last_timestamp=None):
        """
        Postprocessing.
        """
        packet = super().get_dict()
        utils.update_timestamp(packet, last_timestamp)
        return packet


class BIADataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.BIA: ['0xC4', '0x07']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x41',
                'checksum': '0x0'
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'sequence_number': 1,
                'data_type': 0,
                'stream_data': [
                    {
                        'timestamp': 1788107093,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 82
                    },
                    {
                        'timestamp': 1788114863,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 5
                    },
                    {
                        'timestamp': 1788122630,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 8
                    },
                    {
                        'timestamp': 1788130399,
                        'real': 0,
                        'imaginary': 0,
                        'frequency_index': 54
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["data_type"] = Int(1)
        self._config["payload"]["bia_info"] = Enums(1, enum_class=BIAAppInfo)
        self._config["payload"]["timestamp1"] = Int(4)
        self._config["payload"]["real1"] = Int(4, sign=True)
        self._config["payload"]["imaginary1"] = Int(4, sign=True)
        self._config["payload"]["frequency_index1"] = Enums(1, enum_class=BIASweepFrequency)
        self._config["payload"]["timestamp2"] = Int(4)
        self._config["payload"]["real2"] = Int(4, sign=True)
        self._config["payload"]["imaginary2"] = Int(4, sign=True)
        self._config["payload"]["frequency_index2"] = Enums(1, enum_class=BIASweepFrequency)
        self._config["payload"]["timestamp3"] = Int(4)
        self._config["payload"]["real3"] = Int(4, sign=True)
        self._config["payload"]["imaginary3"] = Int(4, sign=True)
        self._config["payload"]["frequency_index3"] = Enums(1, enum_class=BIASweepFrequency)
        self._config["payload"]["timestamp4"] = Int(4)
        self._config["payload"]["real4"] = Int(4, sign=True)
        self._config["payload"]["imaginary4"] = Int(4, sign=True)
        self._config["payload"]["frequency_index4"] = Enums(1, enum_class=BIASweepFrequency)

    def get_dict(self, last_timestamp=None):
        """
        Reorganising of stream data in stream_data key.
        """
        packet = super().get_dict()
        packet["payload"]["stream_data"] = []
        for i in range(1, 5):
            timestamp = packet["payload"][f"timestamp{i}"]
            data = {"timestamp": timestamp,
                    "real": packet["payload"][f"real{i}"],
                    "imaginary": packet["payload"][f"imaginary{i}"],
                    "frequency_index": packet["payload"][f"frequency_index{i}"]}
            [packet["payload"].pop(key) for key in
             [f"timestamp{i}", f"real{i}", f"imaginary{i}", f"frequency_index{i}"]]
            packet["payload"]["stream_data"].append(data)
        utils.update_timestamp(packet, last_timestamp)
        return packet


class AGCDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.AGC: ['0xC6', '0xB0']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 48,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.NEW_STREAM_STATUS: ['0x43']>,
                'sequence_number': 0,
                'timestamp': 1633077776386.271,
                'mts': [ 65535, 0, 0, 0, 0, 0 ],
                'setting': [ 7, 83, 0, 320, 58331, 50, 0, 0, 0, 0 ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["mts"] = Array(12, dimension=1, data_types=[Int(2)])
        self._config["payload"]["setting"] = Array(20, dimension=1, data_types=[Int(2)])

    def get_dict(self, last_timestamp=None):
        """
        Postprocessing.
        """
        packet = super().get_dict()
        utils.update_timestamp(packet, last_timestamp)
        return packet


class HRVDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Stream.HRV: ['0xC6', '0xC0']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 46,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.STREAM_DATA: ['0x28']>,
                'status': <CommonStatus.NEW_STREAM_STATUS: ['0x43']>,
                'sequence_number': 0,
                'stream_data': [
                    {
                        'timestamp': 1633077488525.1702,
                        'rr_interval': 0,
                        'is_gap': 0,
                        'rmssd': 0
                    },
                    {
                        'timestamp': 1633077488527.2014,
                        'rr_interval': 0,
                        'is_gap': 0,
                        'rmssd': 0
                    },
                    {
                        'timestamp': 1633077488529.1702,
                        'rr_interval': 0,
                        'is_gap': 0,
                        'rmssd': 0
                    },
                    {
                        'timestamp': 1633077488531.2017,
                        'rr_interval': 0,
                        'is_gap': 0,
                        'rmssd': 0
                    }
                ]
            }
        }
    """

    def __init__(self):
        super().__init__()
        self._config["payload"]["sequence_number"] = Int(2)
        self._config["payload"]["timestamp1"] = Int(4)
        self._config["payload"]["rr_interval1"] = Int(2, sign=True)
        self._config["payload"]["is_gap1"] = Int(2)
        self._config["payload"]["rmssd1"] = Int(2)
        self._config["payload"]["timestamp2"] = Int(2)
        self._config["payload"]["rr_interval2"] = Int(2, sign=True)
        self._config["payload"]["is_gap2"] = Int(2)
        self._config["payload"]["rmssd2"] = Int(2)
        self._config["payload"]["timestamp3"] = Int(2)
        self._config["payload"]["rr_interval3"] = Int(2, sign=True)
        self._config["payload"]["is_gap3"] = Int(2)
        self._config["payload"]["rmssd3"] = Int(2)
        self._config["payload"]["timestamp4"] = Int(2)
        self._config["payload"]["rr_interval4"] = Int(2, sign=True)
        self._config["payload"]["is_gap4"] = Int(2)
        self._config["payload"]["rmssd4"] = Int(2)

    def get_dict(self, last_timestamp=None):
        """
        Postprocessing.
        """
        packet = super().get_dict()
        packet["payload"]["stream_data"] = []
        timestamp = 0
        for i in range(1, 5):
            timestamp += packet["payload"][f"timestamp{i}"]
            data = {
                "timestamp": timestamp,
                "rr_interval": packet["payload"][f"rr_interval{i}"],
                "is_gap": packet["payload"][f"is_gap{i}"],
                "rmssd": packet["payload"][f"rmssd{i}"],
            }
            [packet["payload"].pop(key) for key in
             [f"timestamp{i}", f"rr_interval{i}", f"is_gap{i}", f"rmssd{i}"]]
            packet["payload"]["stream_data"].append(data)
        utils.update_timestamp(packet, last_timestamp)
        return packet


class KeyStreamDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.DISPLAY: ['0xC5', '0x03']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xB',
                'checksum': '0x0'
            },
            'payload': {
                'command': <DisplayCommand.KEY_STREAM_DATA: ['0x48']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'key_code': 18
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["key_code"] = Int(1)


class CapSenseStreamDataPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.CAP_SENSE_STREAM_DATA: ['0x82']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'position': 1,
                'value': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["position"] = Int(1)
        self._config["payload"]["value"] = Int(1)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import logging
from datetime import datetime

from .common_application import CommonApplication
from ..core.enums.common_enums import CommonCommand, Stream
from ..core.packets.common_packets import StreamPacket, StreamStatusPacket

logger = logging.getLogger(__name__)


class CommonStream(CommonApplication):
    """
    A Common Stream class for streaming data from sensors.
    """

    def __init__(self, destination, stream, packet_manager):
        """
        Initialize the common stream packet variable.

        :param destination: Address of the application.
        :param stream: Address of stream.
        :param packet_manager: PacketManager Object.
        """
        super().__init__(destination, packet_manager)
        self._args = {}
        self._stream = stream
        self._csv_logger = {}
        self._last_timestamp = {}
        self._callback_function = {}

    def set_callback(self, callback_function, args=()):
        """
        Sets the callback for the stream data.

        :param args: optional arguments that will be passed with the callback.
        :param callback_function: callback function for stream adxl data.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 4,12

            from adi_study_watch import SDK

            # make sure optional arguments have default value to prevent them causing Exceptions.
            def callback(data, optional1=None, optional2=None):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            # these optional arguments can be used to pass file, matplotlib or other objects to manipulate data.
            optional_arg1 = "1"
            optional_arg2 = "2"
            application.set_callback(callback, args=(optional_arg1, optional_arg2))
        """
        self._args[self._stream] = args
        self._callback_function[self._stream] = callback_function

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        pass

    def _update_stream_data(self, result):
        """
        Add or modify stream data values.
        """
        pass

    def _callback_data_helper(self, packet, response_packet, stream=None):
        """
        Process and returns the data back to user's callback function.
        """
        stream = stream if stream else self._stream
        args = self._args.get(stream, ())
        callback_function = self._callback_function.get(stream, None)
        response_packet.decode_packet(packet)
        last_timestamp = self._last_timestamp.get(stream)
        result = response_packet.get_dict(last_timestamp)
        self._update_stream_data(result)
        csv_logger = self._csv_logger.get(stream, None)
        if csv_logger:
            csv_logger.add_row(result, last_timestamp[0])
        try:
            if callback_function:
                callback_function(result, *args)
        except Exception as e:
            logger.error(f"Can't send packet back to user callback function, reason :: {e}", exc_info=True)

        if not csv_logger and not callback_function:
            logger.warning(f"No callback function provided for {result['header']['source']}")

    def get_sensor_status(self):
        """
        Returns packet with number of subscribers and number of sensor start request registered.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.get_sensor_status()
            print(x["payload"]["num_subscribers"], x["payload"]["num_start_registered"])
            # 0 0

        """
        request_packet = StreamStatusPacket(self._destination, CommonCommand.GET_SENSOR_STATUS_REQ)
        request_packet.set_payload("stream_address", self._destination)
        response_packet = StreamStatusPacket(self._destination, CommonCommand.GET_SENSOR_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def start_and_subscribe_stream(self, stream=None):
        """
        Starts sensor and also subscribe to the stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            start_sensor, subs_stream = application.start_and_subscribe_stream()
            print(start_sensor["payload"]["status"], subs_stream["payload"]["status"])
            # CommonStatus.STREAM_STARTED CommonStatus.SUBSCRIBER_ADDED
        """
        status2 = self.subscribe_stream(stream=stream)
        status1 = self.start_sensor()
        return status1, status2

    def start_sensor(self):
        """
        Starts sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            start_sensor = application.start_sensor()
            print(start_sensor["payload"]["status"])
            # CommonStatus.STREAM_STARTED
        """
        request_packet = StreamPacket(self._destination, CommonCommand.START_SENSOR_REQ)
        response_packet = StreamPacket(self._destination, CommonCommand.START_SENSOR_RES)
        return self._send_packet(request_packet, response_packet)

    def stop_and_unsubscribe_stream(self, stream=None):
        """
        Stops sensor and also Unsubscribe the stream.

        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            stop_sensor, unsubscribe_stream = application.stop_and_unsubscribe_stream()
            print(stop_sensor["payload"]["status"], unsubscribe_stream["payload"]["status"])
            # CommonStatus.STREAM_STOPPED CommonStatus.SUBSCRIBER_REMOVED
        """
        status1 = self.stop_sensor()
        status2 = self.unsubscribe_stream(stream=stream)
        return status1, status2

    def stop_sensor(self):
        """
        Stops sensor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            stop_sensor = application.stop_sensor()
            print(stop_sensor["payload"]["status"])
            # CommonStatus.STREAM_STOPPED
        """
        request_packet = StreamPacket(self._destination, CommonCommand.STOP_SENSOR_REQ)
        response_packet = StreamPacket(self._destination, CommonCommand.STOP_SENSOR_RES)
        return self._send_packet(request_packet, response_packet)

    def subscribe_stream(self, stream=None):
        """
        Subscribe to the stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            subs_stream = application.subscribe_stream()
            print(subs_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_ADDED
        """
        stream = stream if stream else self._stream
        request_packet = StreamPacket(self._destination, CommonCommand.SUBSCRIBE_STREAM_REQ)
        request_packet.set_payload("stream_address", stream)
        self._subscribe_stream_data(stream=stream)
        response_packet = StreamPacket(self._destination, CommonCommand.SUBSCRIBE_STREAM_RES)
        self._update_timestamp(datetime.now(), stream=stream)
        return self._send_packet(request_packet, response_packet)

    def _update_timestamp(self, date_time, stream=None, generate_ts=False):
        stream = stream if stream else self._stream
        if generate_ts:
            ts = (32000.0 * ((date_time.hour * 3600) + (date_time.minute * 60) + date_time.second))
        else:
            ts = -1
        if stream == Stream.SYNC_PPG:
            self._last_timestamp[stream] = [date_time.timestamp(), ts, date_time.timestamp(), ts]
        else:
            self._last_timestamp[stream] = [date_time.timestamp(), ts]

    def unsubscribe_stream(self, stream=None):
        """
        Unsubscribe the stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            unsubscribe_stream = application.unsubscribe_stream()
            print(unsubscribe_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_REMOVED
        """
        stream = stream if stream else self._stream
        request_packet = StreamPacket(self._destination, CommonCommand.UNSUBSCRIBE_STREAM_REQ)
        request_packet.set_payload("stream_address", stream)
        response_packet = StreamPacket(self._destination, CommonCommand.UNSUBSCRIBE_STREAM_RES)
        response_packet = self._send_packet(request_packet, response_packet)
        self._unsubscribe_stream_data(stream=stream)
        return response_packet

    def _subscribe_stream_data(self, stream=None, callback_function=None):
        stream = stream if stream else self._stream
        callback_function = callback_function if callback_function else self._callback_data
        data_packet_id = self._get_packet_id(CommonCommand.STREAM_DATA, stream)
        self._packet_manager.subscribe(data_packet_id, callback_function)

    def _unsubscribe_stream_data(self, stream=None, callback_function=None):
        stream = stream if stream else self._stream
        callback_function = callback_function if callback_function else self._callback_data
        data_packet_id = self._get_packet_id(CommonCommand.STREAM_DATA, stream)
        self._packet_manager.unsubscribe(data_packet_id, callback_function)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import logging
from typing import Dict, List, Callable, Tuple

from ..core import utils
from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.packets.command_packet import CommandPacket
from ..core.packets.common_packets import VersionPacket
from ..core.enums.ppg_enums import PPGLcfgId, PPGCommand
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.packets.stream_data_packets import PPGDataPacket, SYNCPPGDataPacket, AGCDataPacket, HRVDataPacket
from ..core.packets.ppg_packets import PPGDCBPacket, PPGLibraryConfigPacket, LibraryConfigDataPacket, \
    SetLibraryConfigPacket

logger = logging.getLogger(__name__)


class PPGApplication(CommonStream):
    """
    PPG Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_ppg_application()

    """

    LCFG_ID_ADPD107 = PPGLcfgId.LCFG_ID_ADPD107
    LCFG_ID_ADPD185 = PPGLcfgId.LCFG_ID_ADPD185
    LCFG_ID_ADPD108 = PPGLcfgId.LCFG_ID_ADPD108
    LCFG_ID_ADPD188 = PPGLcfgId.LCFG_ID_ADPD188
    LCFG_ID_ADPD4000 = PPGLcfgId.LCFG_ID_ADPD4000

    STREAM_PPG = Stream.PPG
    STREAM_SYNC_PPG = Stream.SYNC_PPG
    STREAM_DYNAMIC_AGC = Stream.DYNAMIC_AGC_STREAM
    STREAM_HRV = Stream.HRV

    def __init__(self, packet_manager):
        super().__init__(Application.PPG, Stream.PPG, packet_manager)
        self._dcb_size = 53

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        PPG Callback.
        """
        stream = Stream(packet[:2])
        response_packet = PPGDataPacket()
        if stream == self.STREAM_PPG:
            response_packet = PPGDataPacket()
        elif stream == self.STREAM_SYNC_PPG:
            response_packet = SYNCPPGDataPacket()
        elif stream == self.STREAM_DYNAMIC_AGC:
            response_packet = AGCDataPacket()
        elif stream == self.STREAM_HRV:
            response_packet = HRVDataPacket()
        self._callback_data_helper(packet, response_packet, stream)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes PPG Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            application.delete_device_configuration_block()
        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = PPGDCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def get_library_configuration(self) -> Dict:
        """
        Returns entire library configuration PPG.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.get_library_configuration()
            print(x["payload"]["data"])
            # [192, 0, 0, 0, 32, 0, 0, 0, 1, 0, ... , 0,0,0]

        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_LCFG_REQ)
        response_packet = LibraryConfigDataPacket(self._destination, CommonCommand.GET_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    @staticmethod
    def get_supported_lcfg_ids() -> List[PPGLcfgId]:
        """
        List all supported lcfg ID for PPG.

        :return: Array of lcfg ID enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.get_supported_lcfg_ids()
            print(x)
            # [<PPGLcfgId.LCFG_ID_ADPD107: ['0x6B']>, ... , <PPGLcfgId.LCFG_ID_ADPD4000: ['0x28']>]
        """
        return [PPGApplication.LCFG_ID_ADPD107, PPGApplication.LCFG_ID_ADPD185, PPGApplication.LCFG_ID_ADPD108,
                PPGApplication.LCFG_ID_ADPD188, PPGApplication.LCFG_ID_ADPD4000]

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
            # []

        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = PPGDCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        response_dict = self._send_packet(request_packet, response_packet)
        response_dict["payload"]["data"] = utils.add_index_to_array(response_dict["payload"]["data"], to_hex=True)
        return response_dict

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x34

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x0']]

        """
        data = [[field, 0] for field in fields]
        request_packet = PPGLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = PPGLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def set_library_configuration(self, lcfg_id: PPGLcfgId) -> Dict:
        """
        Set PPG to specified library configuration.

        :param lcfg_id: PPG lcfg_id to set, use get_supported_lcfg_ids() to list all supported lcfg IDs
        :type lcfg_id: PPGLcfgId
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.set_library_configuration(application.LCFG_ID_ADPD4000)
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        request_packet = SetLibraryConfigPacket(self._destination, CommonCommand.SET_LCFG_REQ)
        request_packet.set_payload("lcfg_id", lcfg_id)
        response_packet = CommandPacket(self._destination, CommonCommand.SET_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def set_callback(self, callback_function: Callable, args: Tuple = (), stream: Stream = STREAM_PPG) -> None:
        """
        Sets the callback for the stream data.

        :param callback_function: callback function for specified adpd stream.
        :param args: optional arguments that will be passed with the callback.
        :param stream: Callback for specified stream.
        :type stream: Stream
        :return: None

        .. code-block:: python3
            :emphasize-lines: 4,12

            from adi_study_watch import SDK

            # make sure optional arguments have default value to prevent them causing Exceptions.
            def callback(data, optional1=None, optional2=None):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            # these optional arguments can be used to pass file, matplotlib or other objects to manipulate data.
            optional_arg1 = "1"
            optional_arg2 = "2"
            application.set_callback(callback, args=(optional_arg1, optional_arg2), stream=application.PPG)
        """
        stream = self._ppg_stream_helper(stream)
        self._callback_function[stream] = callback_function
        self._args[stream] = args

    def subscribe_stream(self, stream=STREAM_PPG) -> Dict:
        """
        Subscribe to the PPG and SYNC PPG stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            subs_stream = application.subscribe_stream()
            print(subs_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_ADDED
        """
        stream = self._ppg_stream_helper(stream)
        return super().subscribe_stream(stream)

    def unsubscribe_stream(self, stream=STREAM_PPG) -> Dict:
        """
        Unsubscribe the PPG and SYNC PPG stream.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            unsubscribe_stream = application.unsubscribe_stream()
            print(unsubscribe_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_REMOVED
        """
        stream = self._ppg_stream_helper(stream)
        return super().unsubscribe_stream(stream)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x34

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.write_device_configuration_block([[0x20, 2], [0x21, 0x1]])
            print(x["payload"]["size"])
            # 2
        """
        dcb_array = [0] * self._dcb_size
        for address_value in addresses_values:
            address = address_value[0]
            value = address_value[1]
            dcb_array[address] = value
        request_packet = PPGDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        request_packet.set_payload("size", len(dcb_array))
        request_packet.set_payload("data", dcb_array)
        response_packet = PPGDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x34

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            application.write_device_configuration_block_from_file("ppg_dcb.lcfg")
        """
        result = self.device_configuration_file_to_list(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x34

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.write_library_configuration([[0x00, 0x1]])
            print(x["payload"]["data"])
            # [['0x0', '0x1']]

        """
        request_packet = PPGLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        request_packet.set_payload("size", len(fields_values))
        request_packet.set_payload("data", fields_values)
        response_packet = PPGLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    @staticmethod
    def get_supported_streams() -> List[Stream]:
        """
        List all supported streams for PPG.

        :return: Array of stream ID enums.
        :rtype: List[Stream]
        """
        return [PPGApplication.STREAM_PPG, PPGApplication.STREAM_SYNC_PPG, PPGApplication.STREAM_DYNAMIC_AGC,
                PPGApplication.STREAM_HRV]

    def _ppg_stream_helper(self, stream: Stream) -> Stream:
        """
        Confirms stream is from list of Enums.
        """
        if stream in self.get_supported_streams():
            return stream
        else:
            logger.warning(f"{stream} is not supported stream, choosing {self.get_supported_streams()[0]} "
                           f"as default stream. use get_supported_streams() to know all supported streams.")
            return self.get_supported_streams()[0]

    def get_version(self) -> Dict:
        """
        Returns PPG version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # ECG_App
            print(x["payload"]["build_version"])
            # TEST ECG_VERSION STRING
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        response_packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def get_algo_version(self) -> Dict:
        """
        Returns PPG version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.get_algo_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # ECG_App
            print(x["payload"]["build_version"])
            # TEST ECG_VERSION STRING
        """
        request_packet = CommandPacket(self._destination, PPGCommand.GET_ALGO_VENDOR_VERSION_REQ)
        response_packet = VersionPacket(self._destination, PPGCommand.GET_ALGO_VENDOR_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_csv_logging(self, filename, header=None, stream: Stream = STREAM_PPG) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :param stream: PPG or SYNC_PPG stream.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.enable_csv_logging("ppg.csv", stream = application.PPG)
        """
        stream = self._ppg_stream_helper(stream)
        if header is None:
            if stream == self.STREAM_PPG:
                header = ["Timestamp", "HR", "Confidence", "HR Type"]
            elif stream == self.STREAM_SYNC_PPG:
                header = ["PPG Timestamp", "PPG", "ADXL Timestamp", "X", "Y", "Z"]
            elif stream == self.STREAM_DYNAMIC_AGC:
                header = ["Timestamp", ] + [f"MTS{i}" for i in range(1, 6)] + [f"Setting{i}" for i in range(1, 11)]
            elif stream == self.STREAM_HRV:
                header = ["Timestamp", "RR Interval", "Is Gap", "RMSSD"]
        self._csv_logger[stream] = CSVLogger(filename, header)

    def disable_csv_logging(self, stream: Stream = STREAM_PPG) -> None:
        """
        Stops logging stream data into CSV.

        :param stream: PPG or SYNC_PPG stream.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ppg_application()
            x = application.disable_csv_logging(stream = application.PPG)
        """
        stream = self._ppg_stream_helper(stream)
        if self._csv_logger.get(stream):
            self._csv_logger[stream].stop_logging()
        self._csv_logger[stream] = None



class DataType:

    def __init__(self, size, reverse, default):
        self._size = size
        self._value = default
        self._reverse = reverse

    def get_size(self):
        return self._size

    def set_value(self, value):
        self._value = value

    def get_value(self):
        return self._value



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from ..data_types.array import Array
from ..data_types.enums import Enums
from ..data_types.integer import Int
from ..data_types.string import String
from ..enums.common_enums import Stream
from .command_packet import CommandPacket
from ..enums.fs_enums import FileType, FSLogging, FSSubState, FSStatus


class BadBlockPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xE',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.GET_BAD_BLOCKS_RES: ['0x7F']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'bad_blocks': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["bad_blocks"] = Int(4)


class ConfigFilePacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["length"] = Int(2)
        self._config["payload"]["bytes"] = Array(70, data_types=[Int(1)])
        self._config["payload"]["status"] = Enums(1, enum_class=FSStatus)


class DebugInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x2A',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.GET_DEBUG_INFO_RES: ['0x89']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'head_pointer': 374,
                'tail_pointer': 4,
                'usb_avg_tx_time': 0,
                'usb_avg_port_write_time': 0,
                'page_read_time': 0,
                'page_write_time': 1
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["head_pointer"] = Int(4)
        self._config["payload"]["tail_pointer"] = Int(4)
        self._config["payload"]["usb_avg_tx_time"] = Int(4)
        self._config["payload"]["usb_avg_port_write_time"] = Int(4)
        self._config["payload"]["page_read_time"] = Int(4)
        self._config["payload"]["init_circular_buffer_flag"] = Int(2)
        self._config["payload"]["mem_full_flag"] = Int(2)
        self._config["payload"]["data_offset"] = Int(2)
        self._config["payload"]["config_file_occupied"] = Int(2)
        self._config["payload"]["page_write_time"] = Int(4)


class FileCountPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xC',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.GET_NUMBER_OF_FILE_RES: ['0x71']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'file_count': 2
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["file_count"] = Int(2)


class FileInfoRequestPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x26',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.GET_FILE_INFO_RES: ['0x8D']>,
                'status': <FSStatus.OK: ['0x41']>,
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["file_index"] = Int(1)


class FileInfoResponsePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x26',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.GET_FILE_INFO_RES: ['0x8D']>,
                'status': <FSStatus.OK: ['0x41']>,
                'filename': '03123EBD.LOG',
                'start_page': 375,
                'end_page': 375,
                'file_size': 426
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["filename"] = String(16)
        self._config["payload"]["start_page"] = Int(4)
        self._config["payload"]["end_page"] = Int(4)
        self._config["payload"]["file_size"] = Int(4)


class FSStreamStatusPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xD',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.GET_STREAM_SUB_STATUS_RES: ['0x53']>,
                'status': <FSStatus.OK: ['0x41']>,
                'stream_address': <Stream.ADXL: ['0xC2', '0x02']>,
                'sub_state': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = Enums(2, enum_class=Stream, reverse=True)
        self._config["payload"]["sub_state"] = Enums(1, enum_class=FSSubState, reverse=True)


class KeyValuePairPacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["value_id"] = Int(16)


class KeyValuePairResponsePacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["key_id"] = Int(2)
        self._config["payload"]["value_id"] = String(16)
        self._config["payload"]["year"] = Int(2)
        self._config["payload"]["month"] = Int(1)
        self._config["payload"]["day"] = Int(1)
        self._config["payload"]["hour"] = Int(1)
        self._config["payload"]["minute"] = Int(1)
        self._config["payload"]["second"] = Int(1)
        self._config["payload"]["tz_sec"] = Int(4)


class LoggingPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xA',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.START_LOGGING_RES: ['0x77']>,
                'status': <FSStatus.OK: ['0x41']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["logging_type"] = Enums(1, enum_class=FSLogging)


class LSRequestPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x37',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.LS_RES: ['0x49']>,
                'status': <FSStatus.OK: ['0x41']>,
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["dir_path"] = Int(2, default=257)


class LSResponsePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x37',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.LS_RES: ['0x49']>,
                'status': <FSStatus.OK: ['0x41']>,
                'filename': '03043B06.LOG',
                'filetype': <FileType.DATA_FILE: ['0x01']>,
                'file_size': 160309
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["filename"] = String(40)
        self._config["payload"]["filetype"] = Enums(1, enum_class=FileType)
        self._config["payload"]["file_size"] = Int(4)


class PageInfoRequestPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x7A',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.PAGE_READ_TEST_RES: ['0x8F']>,
                'status': <FSStatus.OK: ['0x41']>,
                'page_num': 300,
                'num_bytes': 10
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["page_num"] = Int(4)
        self._config["payload"]["num_bytes"] = Int(1)


class PageInfoResponsePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x7A',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.PAGE_READ_TEST_RES: ['0x8F']>,
                'status': <FSStatus.OK: ['0x41']>,
                'page_num': 300,
                'ecc_zone_status': 0,
                'next_page': 301,
                'occupied': 1,
                'data_region_status': 0,
                'data': [ 4, 0, 74, 71, 104, 6, 75, 126, 4, 0 ],
                'num_bytes': 10
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["page_num"] = Int(4)
        self._config["payload"]["ecc_zone_status"] = Int(1)
        self._config["payload"]["next_page"] = Int(4)
        self._config["payload"]["occupied"] = Int(1)
        self._config["payload"]["data_region_status"] = Int(1)
        self._config["payload"]["data"] = Array(100, data_types=[Int(1)])
        self._config["payload"]["num_bytes"] = Int(1)

    def get_dict(self, last_timestamp=None):
        packet = super().get_dict(last_timestamp)
        packet["payload"]["data"] = packet["payload"]["data"][:packet["payload"]["num_bytes"]]
        return packet


class PatternWritePacket(CommandPacket):

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["file_size"] = Int(4)
        self._config["payload"]["scale_type"] = Int(1)
        self._config["payload"]["scale_factor"] = Int(2)
        self._config["payload"]["num_files_to_write"] = Int(2)


class StreamDebugInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x26',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.STREAM_DEBUG_INFO_RES: ['0x57']>,
                'status': <FSStatus.OK: ['0x41']>,
                'stream_address': <Stream.ADXL: ['0xC2', '0x02']>,
                'packets_received': 0,
                'packets_missed': 0,
                'last_page_read': 0,
                'last_page_read_offset': 0,
                'last_page_read_status': 0,
                'num_bytes_transferred': 0,
                'bytes_read': 0,
                'usb_cdc_write_failed': 0
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["stream_address"] = Enums(2, enum_class=Stream, reverse=True)
        self._config["payload"]["packets_received"] = Int(4)
        self._config["payload"]["packets_missed"] = Int(4)
        self._config["payload"]["last_page_read"] = Int(4)
        self._config["payload"]["last_page_read_offset"] = Int(4)
        self._config["payload"]["last_page_read_status"] = Int(1)
        self._config["payload"]["num_bytes_transferred"] = Int(4)
        self._config["payload"]["bytes_read"] = Int(4)
        self._config["payload"]["usb_cdc_write_failed"] = Int(1)


class StreamFileChunkPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x20E',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.CHUNK_RETRANSMIT_RES: ['0x85']>,
                'status': <FSStatus.OK: ['0x41']>,
                'retransmit_type': 0,
                'page_roll_over': 1,
                'page_chunk_number': 0,
                'page_number': 2,
                'filename': 'FILENAME.LOG'
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["retransmit_type"] = Int(1)
        self._config["payload"]["page_roll_over"] = Int(1)
        self._config["payload"]["page_chunk_number"] = Int(1)
        self._config["payload"]["page_number"] = Int(2)
        self._config["payload"]["filename"] = String(-1)


class StreamFileRequestPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x20E',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.DOWNLOAD_LOG_RES: ['0x7B']>,
                'status': <FSStatus.OK: ['0x41']>,
                'filename': 'FILENAME.LOG'
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["filename"] = String(-1)


class StreamFileResponsePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x20E',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.DOWNLOAD_LOG_RES: ['0x7B']>,
                'status': <FSStatus.OK: ['0x41']>,
                'page_chunk_number': 1,
                'page_number': 0,
                'page_chunk_size': 512,
                'page_chunk_bytes': [ 0, 195, 0, ... , 0, 0, 0, 0, 0, 0, 0 ],
                'crc16': 4312
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["page_chunk_number"] = Int(1)
        self._config["payload"]["page_number"] = Int(2)
        self._config["payload"]["page_chunk_size"] = Int(2)
        self._config["payload"]["page_chunk_bytes"] = Array(512, data_types=[Int(1)])
        self._config["payload"]["crc16"] = Int(2)


class VolumeInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.FS: ['0xC5', '0x01']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x14',
                'checksum': '0x0'
            },
            'payload': {
                'command': <FSCommand.VOL_INFO_RES: ['0x4F']>,
                'status': <FSStatus.OK: ['0x41']>,
                'total_memory': 536870656,
                'used_memory': 487424,
                'available_memory': 99
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["total_memory"] = Int(4)
        self._config["payload"]["used_memory"] = Int(4)
        self._config["payload"]["available_memory"] = Int(2)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import math
import time
import logging
from typing import Dict, List, Union, Tuple, Callable

from tqdm import tqdm

from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.packets.command_packet import CommandPacket
from ..core.packets.common_packets import StreamStatusPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.packets.stream_data_packets import ADPDDataPacket, AGCDataPacket
from ..core.packets.common_packets import VersionPacket, DecimationFactorPacket
from ..core.packets.adpd_packets import ClockCalibrationPacket, ADPDPauseResumePacket
from ..core.enums.adpd_enums import ADPDDevice, ADPDCommand, ADPDSlot, Clock, ADPDLed, ADPDAppID
from ..core.packets.adpd_packets import SamplingFrequencyPacket, ADPDConfigPacket, ADPDUCHRPacket
from ..core.packets.adpd_packets import SlotPacket, ActiveSlotPacket, ComModePacket, AgcControlPacket
from ..core.packets.adpd_packets import ExternalStreamODR, ExternalStreamData, ADPDCreateDeviceConfiguration, ADPDDCFGPacket, \
    ADPDDCBPacket, ADPDRegisterReadPacket, ADPDRegisterWritePacket, ADPDLibraryConfigPacket

logger = logging.getLogger(__name__)


class ADPDApplication(CommonStream):
    """
    ADPD Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_adpd_application()

    """

    STREAM_ADPD1 = Stream.ADPD1
    STREAM_ADPD2 = Stream.ADPD2
    STREAM_ADPD3 = Stream.ADPD3
    STREAM_ADPD4 = Stream.ADPD4
    STREAM_ADPD5 = Stream.ADPD5
    STREAM_ADPD6 = Stream.ADPD6
    STREAM_ADPD7 = Stream.ADPD7
    STREAM_ADPD8 = Stream.ADPD8
    STREAM_ADPD9 = Stream.ADPD9
    STREAM_ADPD10 = Stream.ADPD10
    STREAM_ADPD11 = Stream.ADPD11
    STREAM_ADPD12 = Stream.ADPD12
    STREAM_STATIC_AGC = Stream.STATIC_AGC_STREAM

    DEVICE_GREEN = ADPDDevice.DEVICE_GREEN
    DEVICE_RED = ADPDDevice.DEVICE_RED
    DEVICE_INFRARED = ADPDDevice.DEVICE_INFRARED
    DEVICE_BLUE = ADPDDevice.DEVICE_BLUE
    DEVICE_G_R_IR_B = ADPDDevice.DEVICE_G_R_IR_B

    SLOT_A = ADPDSlot.SLOT_A
    SLOT_B = ADPDSlot.SLOT_B
    SLOT_C = ADPDSlot.SLOT_C
    SLOT_D = ADPDSlot.SLOT_D
    SLOT_E = ADPDSlot.SLOT_E
    SLOT_F = ADPDSlot.SLOT_F
    SLOT_G = ADPDSlot.SLOT_G
    SLOT_H = ADPDSlot.SLOT_H
    SLOT_I = ADPDSlot.SLOT_I
    SLOT_J = ADPDSlot.SLOT_J
    SLOT_K = ADPDSlot.SLOT_K
    SLOT_L = ADPDSlot.SLOT_L

    NO_CLOCK = Clock.NO_CLOCK
    CLOCK_32K = Clock.CLOCK_32K
    CLOCK_1M = Clock.CLOCK_1M
    CLOCK_32M = Clock.CLOCK_32M
    CLOCK_32K_AND_1M = Clock.CLOCK_32K_AND_1M
    CLOCK_1M_AND_32M = Clock.CLOCK_1M_AND_32M

    LED_MWL = ADPDLed.LED_MWL
    LED_GREEN = ADPDLed.LED_GREEN
    LED_RED = ADPDLed.LED_RED
    LED_IR = ADPDLed.LED_IR
    LED_BLUE = ADPDLed.LED_BLUE

    APP_ECG = ADPDAppID.APP_ECG
    APP_PPG = ADPDAppID.APP_PPG
    APP_TEMPERATURE_THERMISTOR = ADPDAppID.APP_TEMPERATURE_THERMISTOR
    APP_TEMPERATURE_RESISTOR = ADPDAppID.APP_TEMPERATURE_RESISTOR
    APP_ADPD_GREEN = ADPDAppID.APP_ADPD_GREEN
    APP_ADPD_RED = ADPDAppID.APP_ADPD_RED
    APP_ADPD_INFRARED = ADPDAppID.APP_ADPD_INFRARED
    APP_ADPD_BLUE = ADPDAppID.APP_ADPD_BLUE

    def __init__(self, packet_manager):
        super().__init__(Application.ADPD, Stream.ADPD6, packet_manager)
        self._dcb_size = 57

    def _adpd_stream_helper(self, stream: Stream) -> Stream:
        """
        Confirms stream is from list of Enums.
        """
        if stream in self.get_supported_streams():
            return stream
        else:
            logger.warning(f"{stream} is not supported stream, choosing {self.get_supported_streams()[5]} "
                           f"as default stream. use get_supported_streams() to know all supported streams.")
            return self.get_supported_streams()[5]

    def calibrate_clock(self, clock_id: Clock) -> Dict:
        """
        Calibrate clock to specified clock ID.

        :param clock_id: Clock ID to calibrate, use get_supported_clocks() to list all supported clock ID.
        :type clock_id: Clock
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5,8

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_clocks()
            print(x)
            # [<Clock.NO_CLOCK: ['0x0']>, ... , <Clock.CLOCK_32K_AND_1M: ['0x5']>, <Clock.CLOCK_1M_AND_32M: ['0x6']>]
            x = application.calibrate_clock(application.CLOCK_1M_AND_32M)
            print(x["payload"]["clock_id"])
            # Clock.CLOCK_1M_AND_32M

        """
        request_packet = ClockCalibrationPacket(self._destination, ADPDCommand.CLOCK_CALIBRATION_REQ)
        request_packet.set_payload("clock_id", clock_id)
        response_packet = ClockCalibrationPacket(self._destination, ADPDCommand.CLOCK_CALIBRATION_RES)
        return self._send_packet(request_packet, response_packet)

    def create_device_configuration(self, slot_app_ids: List[List[Union[ADPDSlot, ADPDAppID]]]) -> Dict:
        """
        Create ADPD device configuration.

        :param slot_app_ids: List of slot ID and APP ID to write, use get_supported_slots() to list all
                            | supported slot ID, and get_supported_app_id() to list all supported app ID.
        :type slot_app_ids: List[List[ADPDSlot, ADPDAppID]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_slots()
            print(x)
            # [<ADPDSlot.SLOT_A: ['0x1']>, <ADPDSlot.SLOT_B: ['0x2']>, ... , <ADPDSlot.SLOT_L: ['0xC']>]
            x = application.get_supported_app_id()
            print(x)
            # [<ADPDAppID.APP_PPG: ['0x1']>, <ADPDAppID.APP_ECG: ['0x0']>, ... , <ADPDAppID.APP_ADPD_BLUE: ['0x7']>]
            x = application.create_device_configuration([[application.SLOT_A, application.APP_ECG],
                                                        [application.SLOT_B, application.APP_ADPD_GREEN]])
            print(x["payload"]["data"])
            # [[<ADPDSlot.SLOT_A: ['0x1']>, <ADPDAppID.APP_ECG: ['0x0']>], ... ]
        """
        request_packet = ADPDCreateDeviceConfiguration(self._destination, ADPDCommand.CREATE_DCFG_REQ)
        request_packet.set_payload("size", len(slot_app_ids))
        request_packet.set_payload("data", slot_app_ids)
        response_packet = ADPDCreateDeviceConfiguration(self._destination, ADPDCommand.CREATE_DCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes ADPD Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            application.delete_device_configuration_block()
        """
        request_packet = CommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = ADPDDCBPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_agc(self, led_list: List[ADPDLed]) -> Dict:
        """
        Disables AGC for LED in list.

        :param led_list: list of led to disable agc, use get_supported_led_ids() to list all supported led ID.
        :type led_list: List[ADPDLed]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_led_ids()
            print(x)
            # [<ADPDLed.LED_GREEN: ['0x0']>, <ADPDLed.LED_RED: ['0x1']>, ... , <ADPDLed.LED_BLUE: ['0x3']>]
            x = application.disable_agc([application.LED_GREEN, application.LED_RED])
            print(x["payload"]["agc_data"])
            # [[<ADPDLed.LED_MWL: ['0x0']>, False], [<ADPDLed.LED_GREEN: ['0x1']>, False]]
        """
        led_list = [[led, 0] for led in led_list]
        request_packet = AgcControlPacket(self._destination, ADPDCommand.AGC_ON_OFF_REQ)
        request_packet.set_payload("size", len(led_list))
        request_packet.set_payload("data", led_list)
        response_packet = AgcControlPacket(self._destination, ADPDCommand.AGC_ON_OFF_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_slot(self, slot_num: ADPDSlot) -> Dict:
        """
        Disable Specified ADPD Slot.

        :param slot_num: slot_num to disable, use get_supported_slots() to list all supported slot ID.
        :type slot_num: ADPDSlot
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_slots()
            print(x)
            # [<ADPDSlot.SLOT_A: ['0x1']>, <ADPDSlot.SLOT_B: ['0x2']>, ... , <ADPDSlot.SLOT_L: ['0xC']>]
            x = application.disable_slot(application.SLOT_A)
            print(x["payload"]["slot_num"], x["payload"]["slot_enabled"])
            # ADPDSlot.SLOT_A False
        """
        request_packet = ActiveSlotPacket(self._destination, ADPDCommand.SET_SLOT_ACTIVE_REQ)
        request_packet.set_payload("slot_num", slot_num)
        request_packet.set_payload("slot_enabled", 0)
        response_packet = ActiveSlotPacket(self._destination, ADPDCommand.SET_SLOT_ACTIVE_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_agc(self, led_list: List[ADPDLed]) -> Dict:
        """
        Enables AGC for LEDs in list.

        :param led_list: list of led to enable agc, use get_supported_led_ids() to list all supported led ID.
        :type led_list: List[ADPDLed]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_led_ids()
            print(x)
            # [<ADPDLed.LED_GREEN: ['0x0']>, <ADPDLed.LED_RED: ['0x1']>, ... , <ADPDLed.LED_BLUE: ['0x3']>]
            x = application.enable_agc([application.LED_GREEN, application.LED_RED])
            print(x["payload"]["agc_data"])
            # [[<ADPDLed.LED_MWL: ['0x0']>, True], [<ADPDLed.LED_GREEN: ['0x1']>, True]]
        """
        led_list = [[led, 1] for led in led_list]
        request_packet = AgcControlPacket(self._destination, ADPDCommand.AGC_ON_OFF_REQ)
        request_packet.set_payload("size", len(led_list))
        request_packet.set_payload("data", led_list)
        response_packet = AgcControlPacket(self._destination, ADPDCommand.AGC_ON_OFF_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_slot(self, slot_num: ADPDSlot) -> Dict:
        """
        Enable Specified ADPD Slot.

        :param slot_num: slot_num to enable, use get_supported_slots() to list all supported slot ID.
        :type slot_num: ADPDSlot
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_slots()
            print(x)
            # [<ADPDSlot.SLOT_A: ['0x1']>, <ADPDSlot.SLOT_B: ['0x2']>, ... , <ADPDSlot.SLOT_L: ['0xC']>]
            x = application.enable_slot(application.SLOT_A)
            print(x["payload"]["slot_num"], x["payload"]["slot_enabled"])
            # ADPDSlot.SLOT_A  True
        """
        request_packet = ActiveSlotPacket(self._destination, ADPDCommand.SET_SLOT_ACTIVE_REQ)
        request_packet.set_payload("slot_num", slot_num)
        request_packet.set_payload("slot_enabled", 1)
        response_packet = ActiveSlotPacket(self._destination, ADPDCommand.SET_SLOT_ACTIVE_RES)
        return self._send_packet(request_packet, response_packet)

    def get_communication_mode(self) -> Dict:
        """
        Get ADPD communication mode.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_communication_mode()
            print(x["payload"]["com_mode"])
            # 2
        """
        request_packet = CommandPacket(self._destination, ADPDCommand.COMMUNICATION_MODE_REQ)
        response_packet = ComModePacket(self._destination, ADPDCommand.COMMUNICATION_MODE_RES)
        return self._send_packet(request_packet, response_packet)

    def get_decimation_factor(self, stream: Stream = STREAM_ADPD6) -> Dict:
        """
        Returns stream decimation factor.

        :param stream: Stream to get decimation factor.
        :type stream: Stream
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD_OPTIONAL: ['0xC2', '0x1D']>]
            x = application.get_decimation_factor(application.STREAM_ADPD6)
            print(x["payload"]["decimation_factor"])
            # 1

        """
        stream = self._adpd_stream_helper(stream)
        request_packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_REQ)
        request_packet.set_payload("stream_address", stream)
        response_packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_RES)
        return self._send_packet(request_packet, response_packet)

    def get_device_configuration(self) -> List[Dict]:
        """
        Returns device configuration data.

        :return: A response packet as dictionary.
        :rtype: List[Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_device_configuration()
            print(x[0]["payload"]["data"])
            # [['0x9', '0x97'], ['0x7', '0x8FFF'], ['0xB', '0x2F6'], ... ]
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_DCFG_REQ)
        response_packet = ADPDDCFGPacket(self._destination, CommonCommand.GET_DCFG_RES)
        return self._send_packet_multi_response(request_packet, response_packet)

    def get_sensor_status(self, stream=STREAM_ADPD6) -> Dict:
        """
        Returns packet with number of subscribers and number of sensor start request registered.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_sensor_status(application.STREAM_ADPD6)
            print(x["payload"]["num_subscribers"], x["payload"]["num_start_registered"])
            # 0 0
        """
        stream = self._adpd_stream_helper(stream)
        request_packet = StreamStatusPacket(self._destination, CommonCommand.GET_SENSOR_STATUS_REQ)
        request_packet.set_payload("stream_address", stream)
        response_packet = StreamStatusPacket(self._destination, CommonCommand.GET_SENSOR_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def get_slot(self, slot_num: ADPDSlot) -> Dict:
        """
        Get Specified ADPD Slot Detail.

        :param slot_num: slot_num to get slot detail, use get_supported_slots() to list all supported slot ID.
        :type slot_num: ADPDSlot
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_slots()
            print(x)
            # [<ADPDSlot.SLOT_A: ['0x1']>, <ADPDSlot.SLOT_B: ['0x2']>, ..., <ADPDSlot.SLOT_L: ['0xC']>]
            x = application.get_slot(application.SLOT_A)
            print(x["payload"]["slot_num"], x["payload"]["slot_enabled"])
            # <ADPDSlot.SLOT_A: ['0x1']> True
            print(x["payload"]["slot_format"], x["payload"]["channel_num"])
            # 3 3
        """
        request_packet = SlotPacket(self._destination, ADPDCommand.GET_SLOT_REQ)
        request_packet.set_payload("slot_num", slot_num)
        response_packet = SlotPacket(self._destination, ADPDCommand.GET_SLOT_RES)
        return self._send_packet(request_packet, response_packet)

    def get_slot_status(self, slot_num) -> Dict:
        """
        Returns whether slot is enabled or not.

        :param slot_num: slot_num to get status, use get_supported_slots() to list all supported slot ID.
        :type slot_num: ADPDSlot
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_slots()
            print(x)
            # [<ADPDSlot.SLOT_A: ['0x1']>, <ADPDSlot.SLOT_B: ['0x2']>, ..., <ADPDSlot.SLOT_L: ['0xC']>]
            x = application.get_slot_status(application.SLOT_A)
            print(x["payload"]["slot_enabled"])
            # True
        """
        request_packet = ActiveSlotPacket(self._destination, ADPDCommand.GET_SLOT_ACTIVE_REQ)
        request_packet.set_payload("slot_num", slot_num)
        response_packet = ActiveSlotPacket(self._destination, ADPDCommand.GET_SLOT_ACTIVE_RES)
        return self._send_packet(request_packet, response_packet)

    @staticmethod
    def get_supported_app_id() -> List[ADPDAppID]:
        """
        List all supported Apps for ADPD.

        :return: Array of APP ID enums.
        :rtype: List[ADPDAppID]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_app_id()
            print(x)
            # [<ADPDAppID.APP_PPG: ['0x1']>, <ADPDAppID.APP_ECG: ['0x0']>, ... , <ADPDAppID.APP_ADPD_BLUE: ['0x7']>]
        """
        return [ADPDApplication.APP_PPG, ADPDApplication.APP_ECG, ADPDApplication.APP_TEMPERATURE_THERMISTOR,
                ADPDApplication.APP_TEMPERATURE_RESISTOR, ADPDApplication.APP_ADPD_GREEN, ADPDApplication.APP_ADPD_RED,
                ADPDApplication.APP_ADPD_INFRARED, ADPDApplication.APP_ADPD_BLUE]

    @staticmethod
    def get_supported_clocks() -> List[Clock]:
        """
        List all supported clocks for ADPD.

        :return: Array of clock ID enums.
        :rtype: List[Clock]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_clocks()
            print(x)
            # [<Clock.NO_CLOCK: ['0x0']>, <Clock.CLOCK_32K: ['0x1']>, ... , <Clock.CLOCK_1M_AND_32M: ['0x6']>]
        """
        return [ADPDApplication.NO_CLOCK, ADPDApplication.CLOCK_32K, ADPDApplication.CLOCK_1M,
                ADPDApplication.CLOCK_32M, ADPDApplication.CLOCK_32K_AND_1M, ADPDApplication.CLOCK_1M_AND_32M]

    @staticmethod
    def get_supported_devices() -> List[ADPDDevice]:
        """
        List all supported devices for ADPD.

        :return: Array of device ID enums.
        :rtype: List[ADPDDevice]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_devices()
            print(x)
            # [<ADPDDevice.DEVICE_GREEN: ['0x28']>, ... , <ADPDDevice.DEVICE_G_R_IR_B: ['0x2C']>]
        """
        return [ADPDApplication.DEVICE_GREEN, ADPDApplication.DEVICE_RED, ADPDApplication.DEVICE_INFRARED,
                ADPDApplication.DEVICE_BLUE, ADPDApplication.DEVICE_G_R_IR_B]

    @staticmethod
    def get_supported_led_ids() -> List[ADPDLed]:
        """
        List all supported led IDs for ADPD.

        :return: Array of Led ID enums.
        :rtype: List[ADPDLed]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_led_ids()
            print(x)
            # [<ADPDLed.LED_GREEN: ['0x1']>, ... , <ADPDLed.LED_BLUE: ['0x4']>, <ADPDLed.LED_MWL: ['0x0']>]
        """
        return [ADPDApplication.LED_GREEN, ADPDApplication.LED_RED, ADPDApplication.LED_IR, ADPDApplication.LED_BLUE,
                ADPDApplication.LED_MWL]

    @staticmethod
    def get_supported_slots() -> List[ADPDSlot]:
        """
        List all supported slots for ADPD.

        :return: Array of slot ID enums.
        :rtype: List[ADPDSlot]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_slots()
            print(x)
            # [<ADPDSlot.SLOT_A: ['0x1']>, <ADPDSlot.SLOT_B: ['0x2']>, ... , <ADPDSlot.SLOT_L: ['0xC']>]
        """
        return [ADPDApplication.SLOT_A, ADPDApplication.SLOT_B, ADPDApplication.SLOT_C, ADPDApplication.SLOT_D,
                ADPDApplication.SLOT_E, ADPDApplication.SLOT_F, ADPDApplication.SLOT_G, ADPDApplication.SLOT_H,
                ADPDApplication.SLOT_I, ADPDApplication.SLOT_J, ADPDApplication.SLOT_K, ADPDApplication.SLOT_L]

    @staticmethod
    def get_supported_streams() -> List[Stream]:
        """
        List all supported streams for ADPD.

        :return: Array of stream ID enums.
        :rtype: List[Stream]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            print(x)
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD12: ['0xC2', '0x1D']>]
        """
        return [ADPDApplication.STREAM_ADPD1, ADPDApplication.STREAM_ADPD2, ADPDApplication.STREAM_ADPD3,
                ADPDApplication.STREAM_ADPD4, ADPDApplication.STREAM_ADPD5, ADPDApplication.STREAM_ADPD6,
                ADPDApplication.STREAM_ADPD7, ADPDApplication.STREAM_ADPD8, ADPDApplication.STREAM_ADPD9,
                ADPDApplication.STREAM_ADPD10, ADPDApplication.STREAM_ADPD11, ADPDApplication.STREAM_ADPD12,
                ADPDApplication.STREAM_STATIC_AGC]

    def get_version(self) -> Dict:
        """
        Returns ADPD version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 0
            print(x["payload"]["minor_version"])
            # 3
            print(x["payload"]["patch_version"])
            # 1
            print(x["payload"]["version_string"])
            # ADPD_App
            print(x["payload"]["build_version"])
            # TEST ADPD4000_VERSION STRING
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        response_packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def load_configuration(self, device_id: ADPDDevice = DEVICE_GREEN) -> Dict:
        """
        Loads specified device id configuration.

        :param device_id: Device ID to load, use get_supported_devices() to list all supported devices.
        :type device_id: ADPDDevice
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5,8

             from adi_study_watch import SDK

             sdk = SDK("COM4")
             application = sdk.get_adpd_application()
             x = application.get_supported_devices()
             print(x)
             # [<ADPDDevice.DEVICE_GREEN: ['0x28']>, ... , <ADPDDevice.DEVICE_G_R_IR_B: ['0x2C']>]
             x = application.load_configuration(application.DEVICE_GREEN)
             print(x["payload"]["device_id"])
             # ADPDDevice.DEVICE_GREEN
         """
        request_packet = ADPDConfigPacket(self._destination, ADPDCommand.LOAD_CONFIG_REQ)
        request_packet.set_payload("device_id", device_id)
        response_packet = ADPDConfigPacket(self._destination, ADPDCommand.LOAD_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def pause(self) -> Dict:
        """
        Pause ADPDDevice.DEVICE_G_R_IR_B.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.pause()
            print(x["payload"]["device_id"], x["payload"]["pause"])
            # ADPDDevice.DEVICE_G_R_IR_B True
        """
        request_packet = ADPDPauseResumePacket(self._destination, ADPDCommand.SET_PAUSE_REQ)
        request_packet.set_payload("device_id", ADPDDevice.DEVICE_G_R_IR_B)
        request_packet.set_payload("pause", 1)
        response_packet = ADPDPauseResumePacket(self._destination, ADPDCommand.SET_PAUSE_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> [Dict]:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: [Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
        """
        request_packet = CommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = ADPDDCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        return self._send_packet_multi_response(request_packet, response_packet)

    def read_library_configuration(self, fields: List[int]) -> Dict:
        """
        Reads library configuration from specified field values.

        :param fields: List of field values to read.
        :type fields: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x00

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.read_library_configuration([0x00])
            print(x["payload"]["data"])
            # [['0x0', '0x12C']]
        """
        data = [[field, 0] for field in fields]
        request_packet = ADPDLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = ADPDLibraryConfigPacket(self._destination, CommonCommand.READ_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_register(self, addresses: List[int]) -> Dict:
        """
        Reads the register values of specified addresses. This function takes a list of addresses to read,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses: List of register addresses to read.
        :type addresses: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x0000
             - 0x0277

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.read_register([0x15, 0x20, 0x2E])
            print(x["payload"]["data"])
            # [['0x15', '0x0'], ['0x20', '0x0'], ['0x2E', '0x0']]
        """
        data = [[address, 0] for address in addresses]
        request_packet = ADPDRegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = ADPDRegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_RES)
        return self._send_packet(request_packet, response_packet)

    def resume(self) -> Dict:
        """
        Resumes ADPDDevice.DEVICE_G_R_IR_B.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.resume()
            print(x["payload"]["device_id"], x["payload"]["pause"])
            # ADPDDevice.DEVICE_G_R_IR_B False
        """
        request_packet = ADPDPauseResumePacket(self._destination, ADPDCommand.SET_PAUSE_REQ)
        request_packet.set_payload("device_id", ADPDDevice.DEVICE_G_R_IR_B)
        request_packet.set_payload("pause", 0)
        response_packet = ADPDPauseResumePacket(self._destination, ADPDCommand.SET_PAUSE_RES)
        return self._send_packet(request_packet, response_packet)

    # noinspection PyTypeChecker
    def set_callback(self, callback_function: Callable, args: Tuple = (), stream: Stream = STREAM_ADPD6) -> None:
        """
        Sets the callback for the stream data.

        :param callback_function: callback function for specified adpd stream.
        :param args: optional arguments that will be passed with the callback.
        :param stream: Callback for specified stream.
        :type stream: Stream
        :return: None

        .. code-block:: python3
            :emphasize-lines: 4,12

            from adi_study_watch import SDK

            # make sure optional arguments have default value to prevent them causing Exceptions.
            def callback(data, optional1=None, optional2=None):
                print(data)

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            # these optional arguments can be used to pass file, matplotlib or other objects to manipulate data.
            optional_arg1 = "1"
            optional_arg2 = "2"
            application.set_callback(callback, args=(optional_arg1, optional_arg2), stream=application.STREAM_ADPD6)
        """
        stream = self._adpd_stream_helper(stream)
        self._callback_function[stream] = callback_function
        self._args[stream] = args

    def set_decimation_factor(self, decimation_factor: int, stream: Stream = STREAM_ADPD6) -> Dict:
        """
        Sets decimation factor for specified ADPD stream.

        :param stream: Stream to set decimation factor.
        :type stream: Stream
        :param decimation_factor: decimation factor for stream
        :type decimation_factor: int
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Decimation Lower Limit
             - Decimation Upper Limit
           * - 0x01
             - 0x05

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD_OPTIONAL: ['0xC2', '0x1D']>]
            x = application.set_decimation_factor(2, application.STREAM_ADPD6)
            print(x["payload"]["decimation_factor"])
            # 2
        """
        stream = self._adpd_stream_helper(stream)
        request_packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_REQ)
        request_packet.set_payload("stream_address", stream)
        request_packet.set_payload("decimation_factor", decimation_factor)
        response_packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_RES)
        return self._send_packet(request_packet, response_packet)

    def set_sampling_frequency(self, odr: int) -> Dict:
        """
        Set ADPD sampling frequency, ODR value in Hz.

        :param odr: ODR frequency in Hz.
        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.set_sampling_frequency(100)
            print(x["payload"]["odr"])
            # 100
        """
        request_packet = SamplingFrequencyPacket(self._destination, ADPDCommand.SET_SAMPLING_FREQUENCY_REQ)
        request_packet.set_payload("odr", odr)
        response_packet = SamplingFrequencyPacket(self._destination, ADPDCommand.SET_SAMPLING_FREQUENCY_RES)
        return self._send_packet(request_packet, response_packet)

    def set_slot(self, slot_num: ADPDSlot, slot_enable: bool, slot_format: int, channel_num: int) -> Dict:
        """
        Set Slot with slot format.

        :param slot_num: slot_num to set slot, use get_supported_slots() to list all supported slot ID.
        :type slot_num: ADPDSlot
        :param slot_enable: enable or disable slot.
        :type slot_enable: bool
        :param slot_format: format of the slot, possible values are 0,1,2,3,4.
        :type slot_format: int
        :param channel_num: channel for the slot, possible values are 1,3.
        :type channel_num: int
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_slots()
            print(x)
            # [<ADPDSlot.SLOT_A: ['0x1']>, <ADPDSlot.SLOT_B: ['0x2']>, ..., <ADPDSlot.SLOT_L: ['0xC']>]
            x = application.set_slot(application.SLOT_A, True, 3, 3)
            print(x["payload"]["slot_num"], x["payload"]["slot_enabled"])
            # ADPDSlot.SLOT_A True
            print(x["payload"]["slot_format"], x["payload"]["channel_num"])
            # 3 3

        """
        if not (channel_num == 0x1 or channel_num == 0x3):
            logger.warning(f"{'0x%X' % channel_num} is out of range, allowed values are: 1,3")
        request_packet = SlotPacket(self._destination, ADPDCommand.SET_SLOT_REQ)
        request_packet.set_payload("slot_num", slot_num)
        request_packet.set_payload("slot_enabled", slot_enable)
        request_packet.set_payload("slot_format", slot_format)
        request_packet.set_payload("channel_num", channel_num)
        response_packet = SlotPacket(self._destination, ADPDCommand.SET_SLOT_RES)
        return self._send_packet(request_packet, response_packet)

    def set_external_stream_sampling_frequency(self, odr: int) -> Dict:
        """
        Set ADPD external stream sampling frequency, ODR value in Hz.

        :param odr: ODR frequency in Hz.
        :return: A response packet as dictionary
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.set_external_stream_sampling_frequency(50)
            print(x["payload"]["odr"])
            # 100
        """
        request_packet = ExternalStreamODR(self._destination, ADPDCommand.SET_EXT_DATA_STREAM_ODR_REQ)
        request_packet.set_payload("odr", odr)
        response_packet = ExternalStreamODR(self._destination, ADPDCommand.SET_EXT_DATA_STREAM_ODR_RES)
        return self._send_packet(request_packet, response_packet)

    def set_external_stream_data(self, csv_filename: str, start_row: int, column_index: int,
                                 display_progress: bool = False) -> None:
        """
        Set csv file data for external adpd stream.

        :param csv_filename: csv file to load stream data.
        :param start_row: start row index of data.
        :param column_index: column index of data
        :param display_progress: display detail progress bar.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            application.set_external_stream_data("12104AD0_ADPDAppStream_SlotFChannel2.csv", 6, 2,
                                                    display_progress=True)
        """
        timestamp = []
        data = []
        try:
            with open(csv_filename, 'r') as file:
                index_count = 1
                line = file.readline()
                csv_time = line.split(',')[2].strip()
                tz_seconds = int(line.split(',')[4].strip())
                hours = int(csv_time[0:2])
                minutes = int(csv_time[3:5])
                seconds = int(csv_time[6:8])
                absolute_time_ms = ((hours * 3600) + (minutes * 60) + seconds + tz_seconds) * 1000
                for line in file:
                    line = line.strip().split(",")
                    if index_count >= start_row:
                        timestamp.append(float(line[0].strip()))
                        data.append(int(line[column_index].strip()))
                    index_count += 1
        except Exception as e:
            logger.error(f"Error while reading the {csv_filename} file, reason :: {e}.", exc_info=True)
            return None
        progress_bar = None
        max_ticks_24_hr = 2764800000

        if display_progress:
            progress_bar = tqdm(total=len(data))
        sequence_number = 1
        for timestamp_value, data_value in zip(timestamp, data):
            request_packet = ExternalStreamData(self._destination, ADPDCommand.EXT_ADPD_DATA_STREAM)
            # converting ms to ticks
            timestamp_value = int(((timestamp_value + absolute_time_ms) * 32.768) % max_ticks_24_hr)
            request_packet.set_payload("sequence_num", sequence_number)
            request_packet.set_payload("data", data_value)
            request_packet.set_payload("timestamp", timestamp_value)
            sequence_number += 1
            time.sleep(0.001)
            if display_progress:
                progress_bar.update(1)
            self._packet_manager.send_packet(request_packet)
        if display_progress:
            progress_bar.close()
        return None

    def start_and_subscribe_stream(self, stream: Stream = STREAM_ADPD6) -> Tuple[Dict, Dict]:
        """
        Starts ADPD sensor and also subscribe to the specified ADPD stream.

        :param stream: Stream to subscribe.
        :type stream: Stream
        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD_OPTIONAL: ['0xC2', '0x1D']>]
            start_sensor, subs_stream = application.start_and_subscribe_stream()
            print(start_sensor["payload"]["status"], subs_stream["payload"]["status"])
            # CommonStatus.STREAM_STARTED CommonStatus.SUBSCRIBER_ADDED
        """
        stream = self._adpd_stream_helper(stream)
        return super().start_and_subscribe_stream(stream)

    def stop_and_unsubscribe_stream(self, stream: Stream = STREAM_ADPD6) -> Tuple[Dict, Dict]:
        """
        Stops ADPD sensor and also Unsubscribe the specified ADPD stream.

        :param stream: Stream to unsubscribe.
        :type stream: Stream
        :return: A response packet as dictionary.
        :rtype: Tuple[Dict, Dict]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD_OPTIONAL: ['0xC2', '0x1D']>]
            stop_sensor, unsubscribe_stream = application.stop_and_unsubscribe_stream()
            print(stop_sensor["payload"]["status"], unsubscribe_stream["payload"]["status"])
            # CommonStatus.STREAM_STOPPED CommonStatus.SUBSCRIBER_REMOVED
        """
        stream = self._adpd_stream_helper(stream)
        return super().stop_and_unsubscribe_stream(stream)

    def subscribe_stream(self, stream: Stream = STREAM_ADPD6) -> Dict:
        """
        Subscribe to the specified ADPD stream.

        :param stream: Stream to subscribe.
        :type stream: Stream
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD_OPTIONAL: ['0xC2', '0x1D']>]
            subs_stream = application.subscribe_stream()
            print(subs_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_ADDED
        """
        stream = self._adpd_stream_helper(stream)
        return super().subscribe_stream(stream)

    def unsubscribe_stream(self, stream: Stream = STREAM_ADPD6) -> Dict:
        """
        Unsubscribe the specified ADPD stream.

        :param stream: Stream to unsubscribe.
        :type stream: Stream
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD_OPTIONAL: ['0xC2', '0x1D']>]
            unsubscribe_stream = application.unsubscribe_stream()
            print(unsubscribe_stream["payload"]["status"])
            # CommonStatus.SUBSCRIBER_REMOVED
        """
        stream = self._adpd_stream_helper(stream)
        return super().unsubscribe_stream(stream)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> [Dict]:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: [Dict]

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x0000
             - 0x0277

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.write_device_configuration_block([[0x20, 2], [0x21, 0x1]])
            print(x["payload"]["size"])
            # 2
        """
        result = []
        packets = math.ceil(len(addresses_values) / self._dcb_size)
        for packet in range(packets):
            addresses_value = addresses_values[packet * self._dcb_size:(packet + 1) * self._dcb_size]
            request_packet = ADPDDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
            request_packet.set_payload("size", len(addresses_value))
            request_packet.set_payload("packet_count", packets)
            request_packet.set_payload("data", addresses_value)
            response_packet = ADPDDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
            result.append(self._send_packet(request_packet, response_packet))
        return result

    def write_device_configuration_block_from_file(self, filename: str) -> [Dict]:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: [Dict]

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x0000
             - 0x0277

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            application.write_device_configuration_block_from_file("adpd4000_dcb.dcfg")
        """
        result = self.device_configuration_file_to_list(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_library_configuration(self, fields_values: List[List[int]]) -> Dict:
        """
        Writes library configuration from List of fields and values.

        :param fields_values: List of fields and values to write.
        :type fields_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Fields Lower Limit
             - Fields Upper Limit
           * - 0x00
             - 0x00

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.write_library_configuration([[0x00, 0x12c]])
            print(x["payload"]["data"])
            # [['0x0', '0x12C']]
        """
        request_packet = ADPDLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_REQ)
        request_packet.set_payload("size", len(fields_values))
        request_packet.set_payload("data", fields_values)
        response_packet = ADPDLibraryConfigPacket(self._destination, CommonCommand.WRITE_LCFG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_register(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the register values of specified addresses. This function takes a list of addresses and values to write,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses_values: List of register addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x0000
             - 0x0277

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.write_register([[0x20, 0x1], [0x21, 0x2], [0x2E, 0x3]])
            print(x["payload"]["data"])
            # [['0x20', '0x1'], ['0x21', '0x2'], ['0x2E', '0x3']]
        """
        request_packet = ADPDRegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = ADPDRegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_RES)
        return self._send_packet(request_packet, response_packet)

    def _callback_data(self, packet, packet_id, callback_function=None, args=None, stream=None):
        """
        Process and returns the data back to user's callback function.
        """
        stream = Stream(packet[:2])
        if stream == self.STREAM_STATIC_AGC:
            self._callback_data_helper(packet, AGCDataPacket(), stream)
        else:
            self._callback_data_helper(packet, ADPDDataPacket(), stream)

    def enable_csv_logging(self, filename, header=None, stream: Stream = STREAM_ADPD6) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :param stream: ADPD Stream.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.enable_csv_logging("adpd6.csv", stream=application.STREAM_ADPD6)
        """
        stream = self._adpd_stream_helper(stream)
        if header is None:
            if stream == self.STREAM_ADPD1:
                header = ["Slot A", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD2:
                header = ["Slot B", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD3:
                header = ["Slot C", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD4:
                header = ["Slot D", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD5:
                header = ["Slot E", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD6:
                header = ["Slot F", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD7:
                header = ["Slot G", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD8:
                header = ["Slot H", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD9:
                header = ["Slot I", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD10:
                header = ["Slot J", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD11:
                header = ["Slot K", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_ADPD12:
                header = ["Slot L", "CH1", "CH2", "Timestamp", "D1", "S1", "D2", "S2"]
            elif stream == self.STREAM_STATIC_AGC:
                header = ["Timestamp", ] + [f"MTS{i}" for i in range(1, 6)] + [f"Setting{i}" for i in range(1, 11)]
        self._csv_logger[stream] = CSVLogger(filename, header, write_header=False)

    def disable_csv_logging(self, stream: Stream = STREAM_ADPD6) -> None:
        """
        Stops logging stream data into CSV.

        :param stream: ADPD Stream.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.disable_csv_logging(stream=application.STREAM_ADPD6)
        """
        stream = self._adpd_stream_helper(stream)
        if self._csv_logger.get(stream):
            self._csv_logger[stream].stop_logging()
        self._csv_logger[stream] = None

    def enable_uc_hr(self, slot: ADPDSlot):
        """
        Enable UC HR.

        :param slot: use get_supported_slots() to list all supported slot ID.
        :type slot: ADPDSlot
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            application.enable_uc_hr(application.SLOT_F)

        """
        request_packet = ADPDUCHRPacket(self._destination, ADPDCommand.UC_HR_ENABLE_REQ)
        request_packet.set_payload("enabled", True)
        request_packet.set_payload("slot", slot)
        response_packet = ADPDUCHRPacket(self._destination, ADPDCommand.UC_HR_ENABLE_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_uc_hr(self, slot: ADPDSlot):
        """
        Disable UC HR.

        :param slot: use get_supported_slots() to list all supported slot ID.
        :type slot: ADPDSlot
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            application.disable_uc_hr(application.SLOT_F)
        """
        request_packet = ADPDUCHRPacket(self._destination, ADPDCommand.UC_HR_ENABLE_REQ)
        request_packet.set_payload("enabled", False)
        request_packet.set_payload("slot", slot)
        response_packet = ADPDUCHRPacket(self._destination, ADPDCommand.UC_HR_ENABLE_RES)
        return self._send_packet(request_packet, response_packet)
